package org.netkernel.http.rest;

import java.io.ByteArrayInputStream;
import java.lang.reflect.Field;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.eclipse.jetty.http.HttpMethod;
import org.netkernel.container.IKernel;
import org.netkernel.container.IMessages;
import org.netkernel.http.util.MimeEtag;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.EndpointMetaImpl;
import org.netkernel.layer0.meta.impl.MetaRepresentationImpl;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.LiteralParameterException;
import org.netkernel.layer0.nkf.impl.NKFContextImpl;
import org.netkernel.layer0.nkf.impl.NKFRequestImpl;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.FastXPath;
import org.netkernel.layer0.util.NullOutputStream;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl.IConfig;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class RESTOverlayConfig implements IConfig
{
	public static final String IGNORE_AMBIGUITY_WARNING = "ignoreAmbiguityWarning";
	private RESTOverlay mRESTOverlay;
	private ISpace mDelegateSpace;
	private ISpaceMeta mDelegateSpaceMeta;
	private String mBasePath;
	private Map<String, LogicalEndpointData> mLogicalEndpoints=new LinkedHashMap<String, LogicalEndpointData>();
	private Map<String, ArrayList<TargetEndpointData>> mGrammarToTargets=new LinkedHashMap<String, ArrayList<TargetEndpointData>>();
	private boolean mStrict;
	private String mAuto404;
	private String mAuto406="";
	private String mAuto404EndpointIdentifier;
	private String mExceptionHandler="";
	private int mGlobalMaxContentLength=-1;
	private ISpaceElements mSpaceElements;
	private Map<String, String> mPassThroughEndpoints=new HashMap<String, String>();
	
	private static final String VERB_TRANSLATE_VERSION1="V1";

	public RESTOverlayConfig(ISpace aDelegateSpace, ISpaceMeta aDelegateSpaceMeta, INKFRequestContext aContext, RESTOverlay aRESTOverlay) throws Exception
	{	mDelegateSpace=aDelegateSpace;
		mDelegateSpaceMeta=aDelegateSpaceMeta;
		mRESTOverlay=aRESTOverlay;
		Document cfg=aContext.source("param:config", Document.class);
		mBasePath=XMLUtils.getText(FastXPath.getSingleNode(cfg.getDocumentElement(), "basepath"));
		mStrict=FastXPath.getSingleNode(cfg.getDocumentElement(), "strict")!=null;
		Node n404=FastXPath.getSingleNode(cfg.getDocumentElement(), "auto404");
		if(n404!=null)
		{	mAuto404=XMLUtils.getText(n404);				
		}
		if(mAuto404==null && n404!=null)
		{	mAuto404="";			
		}
		Node n406=FastXPath.getSingleNode(cfg.getDocumentElement(), "auto406");
		if(n406!=null)
		{	mAuto406=XMLUtils.getText(n406);
		}
		Node nExceptionHandler=FastXPath.getSingleNode(cfg.getDocumentElement(), "globalExceptionTarget");
		if(nExceptionHandler!=null)
		{	mExceptionHandler=XMLUtils.getText(nExceptionHandler);
		}
		Node nMaxContentSize=FastXPath.getSingleNode(cfg.getDocumentElement(), "maxContentLength");
		if(nMaxContentSize!=null)
		{	mGlobalMaxContentLength=Integer.parseInt(XMLUtils.getText(nMaxContentSize));				
		}
		aRESTOverlay.clearInitialisationProblems();
		initialize(aContext, aRESTOverlay);
	}

	public ISpaceElements getElements(INKFRequestContext aContext) throws Exception
	{	return mSpaceElements;
	}

	public IMetaRepresentation getElementMeta(String aElementId)
	{	LogicalEndpointData led=mLogicalEndpoints.get(aElementId);
		return led.mEndpointMeta;
	}

	public String getResolvedElement(INKFResolutionContext aContext) throws Exception
	{	String result=null;
		INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
		for (String key: mLogicalEndpoints.keySet())
		{	IMetaRepresentation meta = mLogicalEndpoints.get(key).mEndpointMeta;
			if (MetadataUtils.match((IEndpointMeta)meta, toResolve, aContext))
			{	result = key;
				break;
			}
		}
		return result;
	}
	
	private boolean isAmbigiousResolution( INKFRequestContext aContext) throws Exception
	{	INKFRequestReadOnly toResolve=aContext.getThisRequest();
		int matchedLogicalEndpoints=0;
		for (String key: mLogicalEndpoints.keySet())
		{	if (!key.equals(mAuto404EndpointIdentifier))
			{	IMetaRepresentation meta = mLogicalEndpoints.get(key).mEndpointMeta;
				if (MetadataUtils.match((IEndpointMeta)meta, toResolve))
				{	matchedLogicalEndpoints++;
				}
			}
		}
		return matchedLogicalEndpoints>1;
	}
	
	private ArrayList<LogicalEndpointData> getLogicalEndpointTargets( INKFRequestContext aContext) throws Exception
	{	INKFRequestReadOnly toResolve=aContext.getThisRequest();
		ArrayList<LogicalEndpointData> r=new ArrayList();
		for (String key: mLogicalEndpoints.keySet())
		{	if (!key.equals(mAuto404EndpointIdentifier))
			{	LogicalEndpointData led=mLogicalEndpoints.get(key);
				IMetaRepresentation meta = led.mEndpointMeta;
				if (MetadataUtils.match((IEndpointMeta)meta, toResolve))
				{	r.add(led);
				}
			}
		}
		return r;
	}
	

	public void onRequest(String aElementId, INKFRequestContext aContext) throws Exception
	{	boolean transientStrict=mStrict;
		if(mAuto404!=null && aElementId.equals(mAuto404EndpointIdentifier))
		{	
			if(mAuto404.equals(""))
			{	INKFResponse resp=aContext.createResponseFrom("404 Resource not found");
				resp.setMimeType("text/plain");
				resp.setHeader("httpResponse:/code", 404);
				return;
			}
			else
			{	INKFRequest req=aContext.createRequestToEndpoint(mAuto404);
				req.addArgumentByValue("request", aContext.getThisRequest());
				INKFResponseReadOnly handlerResp=aContext.issueRequestForResponse(req);
				INKFResponse resp=aContext.createResponseFrom(handlerResp);
				try
				{	handlerResp.getHeader("httpResponse:/code");
				}	
				catch(Exception e2)
				{	resp.setHeader("httpResponse:/code", 404);
				}
				return;
			}
		}
		LogicalEndpointData led=mLogicalEndpoints.get(aElementId);
		if(led.mMappings==null)
		{	//This is a delegated non-REST endpoint so just pass-through
			String targetID=mPassThroughEndpoints.get(aElementId);
			INKFRequest req=aContext.getThisRequest().getIssuableClone();
			
			aContext.createResponseFrom(aContext.issueRequestForResponse(req));
			return;
		}
		ArrayList<TargetEndpointData> mappings=mGrammarToTargets.get(led.mGrammar);  //Must use these targets and not LogicalEndpointData.mMappings since we may have resolved with an ambiguous grammar and now need to demux with REST metadata
		HttpServletRequest sreq;
		try
		{	sreq=aContext.source("httpRequest:/advanced/HttpServletRequest", HttpServletRequest.class);
		}
		catch(Exception e)
		{	throw new NKFException("RESTOverlay cannot find HTTPBridge","RESTOverlay can only be called when HTTPBridge is in the request scope");		
		}
		//Need this a few places so get it here
		String method=sreq.getMethod();
		
		INKFRequest req=null;
		//Sorting arrays
		ArrayList<TargetEndpointData> candidates=new ArrayList(4);
		ArrayList matches=new ArrayList(4);
		ArrayList possibles=new ArrayList(4);
		IHDSNode accepts=aContext.source("httpRequest:/accept", IHDSNode.class);
		String requiredMime=null;
		boolean matchedWithoutAcceptsCompare=false;
		boolean matchedWithAcceptLanguage=false;
		boolean varyWithCompression=false;
		
		boolean sayNoVerb=false;
		boolean sayNoAccept=false;
		boolean sayNoConsume=false;
		boolean sayNoLang=false;
		boolean do405=false;		//When method does not match we need to respond with 405 and Allow verbs.
		
		boolean implementsPATCH=false;
		
		//System.out.println(accepts);
		while(true)
		{	//Options Handling
			if("OPTIONS".equals(method))
			{	String options = getAllowedMethods(mappings);
				INKFResponse resp=aContext.createResponseFrom("200 "+led.mGrammar+"\nOPTIONS: "+options);
				resp.setHeader("httpResponse:/code", 200);
				resp.setHeader("httpResponse:/header/Allow", options);
				if(options.contains("PATCH"))
				{	StringBuilder sb=new StringBuilder();
					for(TargetEndpointData ted: mappings)
					{	if(ted.mMethod.equals("PATCH") && ted.mConsumes!=null)
						{	sb.append(ted.mConsumes).append(",");							
						}
						
					}
					String pc=sb.toString();
					if(pc.length()>0)
					{	pc=pc.substring(0,  pc.length()-1);
						resp.setHeader("httpResponse:/header/Accept-Patch", pc);
					}
				}				
				return;
			}
			//Special case for PATCH - This should eventually be transactional since between verifying GET Etag and applying PATCH someone could have PUT/DELETE/POSTed the resource! 
			if("PATCH".equals(method))
			{	String ifmatch=sreq.getHeader("If-Match");
				if(ifmatch!=null)
				{
					String contentType=sreq.getContentType();
					boolean patchValid=false;
					for(TargetEndpointData ted: mappings)
					{	if(ted.mMethod.equals("GET") && ted.mEtag!=null )
						{	
							String mime=null;
							try
							{	//Attempt to decode originating mime from If-Match - this is a workaround
								//for a serious failure of the PATCH IETF specification!!!
								Map<String,String> map=MimeEtag.decode(ifmatch.replaceAll("\"", "").trim());
								mime=map.get("mime");
								System.out.println("PATCH MIME DECODE "+mime);
							}
							catch(Exception e)
							{	//Not a composite Etag so just go with regular hash comparison.								
							}
							boolean requiredMimeFlag=false;
							if(ted.mProduces.size()>1)
							{	//This target has multiple produces - we need to see if it produces more than one type without an @withTransform - if it does then we need to relay the "requiredMime" argument.
								for(String key: ted.mProduces.keySet())
								{	if(key!=mime)
									{	IHDSNode produces = ted.mProduces.get(key);
										if(produces.getFirstValue("@withTransform")==null)	//Cannot support produces @withTransform as this would explode permutations of PATCH
										{	requiredMimeFlag=true;
											break;
										}
									}
								}
							}
							req=prepareRequest(aContext,ted.mIdentifier, null, null, ted.mMethod, requiredMimeFlag, mime);
							req.setRepresentationClass(IReadableBinaryStreamRepresentation.class);
							IReadableBinaryStreamRepresentation rep=(IReadableBinaryStreamRepresentation)aContext.issueRequest(req);
							
							INKFRequest etagReq=aContext.createRequest("active:httpEtag");
							etagReq.addArgumentByValue("operand", rep);
							etagReq.addArgumentByValue("operator", ted.mEtag);
							if(mime!=null)
							{	etagReq.addArgumentByValue("mime", mime);
							}
							IRequestScopeLevel scope=getWormholeScopeToHTTPTools(aContext);
							etagReq.setRequestScope(scope);
							String eTag=(String)aContext.issueRequest(etagReq);
							
							if(!eTag.contains(ifmatch))
							{	//Report error - this resource cannot be patched as it doesn't match
								/*
								 * Conflicting modification:  When a client uses either the If-Match or
							      If-Unmodified-Since header to define a precondition, and that
							      precondition failed, then the 412 (Precondition Failed) error is
							      most helpful to the client.  However, that response makes no sense
							      if there was no precondition on the request.  In cases when the
							      server detects a possible conflicting modification and no
							      precondition was defined in the request, the server can return a
							      409 (Conflict) response.
								*/
								doXXX(aContext, "412 Cannot PATCH "+led.mGrammar+" local state does not match If-Match etag", 412);
								return;
							}
							else
							{	//candidates.add(ted);
								patchValid=true;
								//Need to always be strict with PATCH
								transientStrict=true;
								break;
							}
						}
					}
					if(!patchValid)
					{	//Couldn't find the If-Match resource to be patched - so must reject
						/*
						Unsupported patch document:  Can be specified using a 415
					      (Unsupported Media Type) response when the client sends a patch
					      document format that the server does not support for the resource
					      identified by the Request-URI.  Such a response SHOULD include an
					      Accept-Patch response header as described in Section 3.1 to notify the
					      client what patch document media types are supported.
					      */
						doXXX(aContext, "415 Cannot PATCH "+led.mGrammar+" with media type "+contentType, 415);
						return;
					}
				}
				//Valid PATCHable Resource is available so can continue as a regular request...
			}
			//Method Matching
			for(TargetEndpointData ted: mappings)
			{	if(ted.mMethod!=null)
				{	if(ted.mMethod.contains(method))
					{	matches.add(ted);					
					}
					if(ted.mMethod.contains("PATCH"))
					{	implementsPATCH=true;
					}
				}
				else if(!transientStrict)
				{	possibles.add(ted);
				}
			}
			filterLists(candidates, matches, possibles);
			if(!transientStrict && candidates.size()==1)
			{	matchedWithoutAcceptsCompare=true;
				break;
			}
			if (candidates.size()==0)
			{	do405=true;
				sayNoVerb=true;
			}
			
			//Accept Matching
			IHDSNodeList nl=accepts.getNodes("/headers/Accept/entry");
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	String type=(String)n.getFirstValue("value");
					Float q=(Float)n.getFirstValue("q");
					int wildcard=type.indexOf('*');
					if(wildcard>=0)
					{	type=type.substring(0, wildcard);						
					}
					for(TargetEndpointData ted: candidates)
					{	if(ted.mProduces!=null)
						{	
							for(String key: ted.mProduces.keySet())
							{	if(key.startsWith(type) && q!=0.0f)
								{	matches.add(ted);
									requiredMime=key;
									break;
								}
							}
						}
						else if(!transientStrict)
						{	possibles.add(ted);
						}
					}
				}
				filterLists(candidates, matches, possibles);
			}
			if(!transientStrict && candidates.size()==1)
			{	break;
			}
			if (candidates.size()==0) sayNoAccept=true;
			
			//Consumes Matching
			if(HttpMethod.POST.is(method) || HttpMethod.PUT.is(method) || method.equals("PATCH"))
			{	String contentType=sreq.getContentType();
				if (contentType!=null)
				{
					int separator=contentType.indexOf(";");
					if(separator>0)
					{	contentType=contentType.substring(0, separator);
					}
					for(TargetEndpointData ted: candidates)
					{	if(!transientStrict && ted.mConsumes==null)
						{	possibles.add(ted);
						}
						else
						{	if(ted.mConsumes!=null && ted.mConsumes.contains(contentType))
							{	matches.add(ted);
								break;
							}
						}
					}
				}
				filterLists(candidates, matches, possibles);
			}
			if(!transientStrict && candidates.size()==1)
			{	break;
			}
			if (candidates.size()==0) sayNoConsume=true;
			
			//Accept-Language  Matching
			nl=accepts.getNodes("/headers/Accept-Language/entry");
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	String type=(String)n.getFirstValue("value");
					Float q=(Float)n.getFirstValue("q");
					for(TargetEndpointData ted: candidates)
					{	if(ted.mLanguage!=null)
						{	if(ted.mLanguage.equals(type)  && q!=0.0f)
							{	matches.add(ted);
								break;
							}
						}
						else if(!transientStrict)
						{	possibles.add(ted);
						}
					}
				}
				if(candidates.size()>0)
				{	matchedWithAcceptLanguage=true;					
				}
				filterLists(candidates, matches, possibles);
				if (candidates.size()==0) sayNoLang=true;
			}
			if(candidates.size()>1)
			{	if(transientStrict)
				{	//Ambiguous 406
					do406(aContext, true, null);
					return;
				}
			}
			if(candidates.size()>=1)
			{	break;
			}
			//Nothing found - either we resolved to wrong ambiguous logical endpoint or there truly is no match
			if (do405)
			{	do405(aContext, mappings);
				return;
			}
			//Check for ambiguous grammars
			
			if (isAmbigiousResolution(aContext))
			{	INKFResponse resp=aContext.createResponseFrom("500 Resource found but conflicting external REST grammars - check <rest><simple>...</simple></rest> they are not unique");
				resp.setMimeType("text/plain");
				resp.setHeader("httpResponse:/code", 500);
				return;
			}
			String msg=null;
			if (sayNoAccept) msg="No Accept header match";
			else if (sayNoConsume) msg="No Consume mime type match";
			else if (sayNoLang) msg="No Accept-Language header match";
			do406(aContext, false,msg);
			return;
		}
		//Issue Request to Target Endpoint
		TargetEndpointData target=candidates.get(0);

		//Do content size check
		if(HttpMethod.POST.is(method) || HttpMethod.PUT.is(method) || method.equals("PATCH"))
		{	int contentLength=sreq.getContentLength();
			boolean toobig=false;
			if(target.mMaxContentLength!=-1)
			{	toobig=contentLength>target.mMaxContentLength  || contentLength==-1;
			}
			else if (mGlobalMaxContentLength!=-1)
			{	toobig=contentLength>mGlobalMaxContentLength || contentLength==-1;;
			}
			if(toobig)
			{	do413(aContext, contentLength==-1);
				return;
			}
		}

		//Move to SSL?
		if(target.mMustBeSSL && !sreq.getProtocol().equals("https"))
		{	INKFResponse resp=aContext.createResponseFrom("302 Redirect - you must use SSL for this service");
			resp.setMimeType("text/plain");
			resp.setHeader("httpResponse:/redirect", sreq.getRequestURL().toString().replaceFirst("http://", "https://"));
			return;
		}
		//Pre Process
		if(target.mPreTarget!=null)
		{	INKFRequest pre=aContext.createRequestToEndpoint(target.mPreTarget);
			pre.addArgumentByValue("request", aContext.getThisRequest());
			pre.setRepresentationClass(Boolean.class);
			Boolean preresp=(Boolean)aContext.issueRequest(pre);
			if(!preresp)
			{	aContext.logRaw(INKFRequestContext.LEVEL_FINE, aContext.getThisRequest().toString()+" did not pass pre process "+target.mPreTarget);
				return;
			}
		}
		
		//What was the required mimetype (for use in sub-request requiredMime argument and Produces withTransform).
		if(matchedWithoutAcceptsCompare)
		{	//Accept Matching
			IHDSNodeList nl=accepts.getNodes("/headers/Accept/entry");
			String firsttype=null;
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	
					String type=(String)n.getFirstValue("value");
					if(firsttype==null)
					{	firsttype=type;						
					}
					Float q=(Float)n.getFirstValue("q");
					if(type.equals("*/*") && q>0.0f)
					{	//They will accept anything so stop looking
						requiredMime=null;
					}
					int wildcard=type.indexOf('*');
					if(wildcard>=0)
					{	type=type.substring(0, wildcard);						
					}
					if(target.mProduces!=null)
					{	for(String key: target.mProduces.keySet())
						{	if(key.startsWith(type) && q!=0.0f)
							{	requiredMime=key;
								break;
							}
						}
					}
					if(requiredMime!=null)
					{	break;
					}
				}
			}
			if(requiredMime==null)
			{	requiredMime=firsttype;  //Default to first match or null.				
			}
		}
		
		//Deal with requiredMime argument for endpoints with multiple produces.
		String withTransform=null;
		boolean setOverlayAcceptArgument = false;
		if(target.mProduces!=null && target.mProduces.size()>1)
		{	IHDSNode produces=target.mProduces.get(requiredMime);
			boolean searchForProduces=true;
			if(produces!=null)
			{	withTransform=(String)produces.getFirstValue("@withTransform");
			}
			else if(!transientStrict)
			{	//This target doesn't produce the desired mimetype - but we're tolerant so it has to accept what its given.  No need for requiredMime argument
				searchForProduces=false;
			}
			
			if(searchForProduces && withTransform==null)
			{	//OK this produces our target mime but is this the only <produces> with no @withTransform - if not we need to disambiguate?
				for(String key: target.mProduces.keySet())
				{	if(key!=requiredMime)
					{	produces = target.mProduces.get(key);
						if(produces.getFirstValue("@withTransform")==null)
						{	setOverlayAcceptArgument=true;  //This endpoint's rep will not be transrepted @withTransform so we need to provide requiredMime.
							break;
						}
					}
				}
			}
		}
		
		//Prepare and Issue Request
		req=prepareRequest(aContext,target.mIdentifier, target.mWrapperTarget, target.mVerbTranslate, method, setOverlayAcceptArgument, requiredMime);
		Object resp=null;
		try
		{	resp=aContext.issueRequestForResponse(req);
		}
		catch(Exception e)
		{	//Exception Handling
			if(target.mExceptionTarget!=null)
			{	doExceptionHandling(aContext, e, target.mExceptionTarget);			
			}
			else if(!mExceptionHandler.equals(""))
			{	doExceptionHandling(aContext, e, mExceptionHandler);
			}
			else
			{	throw e;				
			}
			return;
		}
		
		//Do the @withTransform we discovered earlier 
		if(withTransform!=null)
		{	INKFResponseReadOnly r=(INKFResponseReadOnly)resp;
			if(!r.getMimeType().equals(requiredMime))
			{	req=aContext.createRequest(withTransform);
				req.addArgumentByValue("operand", r.getRepresentation());
				resp=aContext.issueRequestForResponse(req);
			}
		}
		
		//Post Process Compression - take care if verb translating method is POST or PATCH as could be inconsistent
		if(target.mCompress && (HttpMethod.GET.is(method) || HttpMethod.POST.is(method) || method.equals("PATCH") ) )
		{	IHDSNodeList nl=accepts.getNodes("/headers/Accept-Encoding/entry");
			if(nl.size()>0)
			{	for(IHDSNode n: nl)
				{	String type=(String)n.getFirstValue("value");
					Float q=(Float)n.getFirstValue("q");
					if(type.equals("gzip")  && q!=0.0f)
					{	IRequestScopeLevel scope=getWormholeScopeToHTTPTools(aContext);
						INKFResponseReadOnly roresp=(INKFResponseReadOnly)resp;
						INKFRequest compressRequest=aContext.createRequest("active:httpGZIPCompress");
						compressRequest.setRequestScope(scope);
						compressRequest.addArgumentFromResponse("operand", roresp);
						INKFResponseReadOnly cresp=aContext.issueRequestForResponse(compressRequest);
						INKFResponse compressedresp=aContext.createResponseFrom(cresp);
						relayHeaders(roresp, compressedresp);
						compressedresp.setHeader("httpResponse:/header/Content-Encoding", "gzip");
						resp=compressedresp;
						varyWithCompression=true;
						break;
					}
					else if(type.equals("deflate") && q!=0.0f)
					{	IRequestScopeLevel scope=getWormholeScopeToHTTPTools(aContext);
						INKFResponseReadOnly roresp=(INKFResponseReadOnly)resp;
						INKFRequest compressRequest=aContext.createRequest("active:httpDeflateCompress");
						compressRequest.setRequestScope(scope);
						compressRequest.addArgumentFromResponse("operand", roresp);
						INKFResponseReadOnly cresp=aContext.issueRequestForResponse(compressRequest);
						INKFResponse compressedresp=aContext.createResponseFrom(cresp);
						relayHeaders(roresp, compressedresp);
						compressedresp.setHeader("httpResponse:/header/Content-Encoding", "deflate");
						resp=compressedresp;
						varyWithCompression=true;
						break;
					}
				}
			}
		}
		//Etag
		/*
		MD5, MD2, SHA-1,SHA-256,SHA-384,SHA-512
		 */
		if(HttpMethod.GET.is(method) && target.mEtag!=null)
		{	INKFResponse finalResponse;
			Object finalRep;
			if(resp instanceof INKFResponse)
			{	finalResponse=(INKFResponse)resp;
				finalRep=finalResponse.getRepresentation();
			}
			else
			{	finalResponse=aContext.createResponseFrom(resp);
				finalRep=((INKFResponseReadOnly)resp).getRepresentation();
			}
			if(finalRep instanceof IReadableBinaryStreamRepresentation)
			{	String mime=finalResponse.getMimeType();
				INKFRequest etagReq=aContext.createRequest("active:httpEtag");
				etagReq.addArgumentByValue("operand", finalRep);
				etagReq.addArgumentByValue("operator", target.mEtag);
				if(implementsPATCH && mime!=null)
				{	etagReq.addArgumentByValue("mime", mime);				
				}
				IRequestScopeLevel scope=getWormholeScopeToHTTPTools(aContext);
				etagReq.setRequestScope(scope);
				String eTag=(String)aContext.issueRequest(etagReq);
				finalResponse.setHeader("httpResponse:/header/Etag", eTag);
				resp=finalResponse;
			}
		}
		
		//Create the response
		INKFResponse finalResponse=aContext.createResponseFrom(resp);

		//Finally if we have used any of Accept, Accept-Language or Accept-Encoding
		//provide Vary: Accept header so that external caches are aware the Acceptxxx headers are significant.
		StringBuffer sb=new StringBuffer();
		if(target.mProduces!=null && target.mProduces.size()>1)
		{	sb.append("Accept,");
		}
		if(matchedWithAcceptLanguage)
		{	sb.append("Accept-Language,");
		}
		if(varyWithCompression)
		{	sb.append("Accept-Encoding,");			
		}
		String vary=sb.toString();
		if(!vary.equals(""))
		{	vary=vary.substring(0, vary.length()-1);	//Trim trailing comma
			finalResponse.setHeader("httpResponse:/header/Vary", vary);
		}
	}

	private String getAllowedMethods(ArrayList<TargetEndpointData> mappings)
	{	StringBuilder sb=new StringBuilder();
		for(TargetEndpointData ted: mappings)
		{	sb.append(ted.mMethod).append(",");
		}
		String options=sb.toString();
		options=options.substring(0, options.length()-1);
		return options;
	}
	
	private void doExceptionHandling(INKFRequestContext aContext, Exception e, String handler) throws Exception
	{	INKFRequest req=aContext.createRequestToEndpoint(handler);
		req.addArgumentByValue("request", aContext.getThisRequest());
		req.addArgumentByValue("exception", new RESTOverlayWrappedException(e));
		INKFResponseReadOnly handlerResp=aContext.issueRequestForResponse(req);
		INKFResponse resp=aContext.createResponseFrom(handlerResp);
		try
		{	handlerResp.getHeader("httpResponse:/code");
		}	
		catch(Exception e2)
		{	resp.setHeader("httpResponse:/code", 500);
		}
		return;
	}
	
	private void do406(INKFRequestContext aContext, boolean ambiguous, String aDetailMsg) throws Exception
	{	if(mAuto406.equals(""))
		{	String message;
			if(ambiguous)
			{ message="406 Resource found but more than one possible endpoint so not acceptable in strict mode";
			}
			else
			{	message="406 Resource found but not acceptable";
				if (aDetailMsg!=null)
				{	message=message+" - "+aDetailMsg;
				}
			}
			INKFResponse resp=aContext.createResponseFrom(message);
			resp.setMimeType("text/plain");
			resp.setHeader("httpResponse:/code", 406);
			return;
		}
		else
		{	INKFRequest req=aContext.createRequestToEndpoint(mAuto406);
			req.addArgumentByValue("request", aContext.getThisRequest());
			INKFResponseReadOnly handlerResp=aContext.issueRequestForResponse(req);
			INKFResponse resp=aContext.createResponseFrom(handlerResp);
			resp.setHeader("httpResponse:/code", 406);
			return;
		}
	}
	
	private void doXXX(INKFRequestContext aContext, String message, int code) throws Exception
	{	INKFResponse resp=aContext.createResponseFrom(message);
		resp.setMimeType("text/plain");
		resp.setHeader("httpResponse:/code", code);
	return;
		
	}
	
	private void do413(INKFRequestContext aContext, boolean notReported) throws Exception
	{	String message;
		if(notReported)
		{	message="413 Content-Length header not reported. Must be set to check acceptable maximum permitted size";
		}
		else
		{	message="413 Content size exceeds maximum permitted";
		}
		INKFResponse resp=aContext.createResponseFrom(message);
		resp.setMimeType("text/plain");
		resp.setHeader("httpResponse:/code", 413);
		return;
	}
	
	//405 Method not allowed
	private void do405(INKFRequestContext aContext, ArrayList<TargetEndpointData> mappings) throws Exception
	{	String message;
		message="405 Method not allowed";
		INKFResponse resp=aContext.createResponseFrom(message);
		resp.setMimeType("text/plain");
		String methods=this.getAllowedMethods(mappings);
		resp.setHeader("httpResponse:/header/Allow", methods);
		resp.setHeader("httpResponse:/code", 405);
		return;
	}
	
	private void relayHeaders(INKFResponseReadOnly oldresp, INKFResponse newresp)
	{	IRequestResponseFields fields=oldresp.getHeaders();
		for(int i=0; i< fields.size(); i++)
		{	String key=fields.getKey(i);
			Object value=fields.getValue(i);
			newresp.setHeader(key, value);
		}
	}
	
	private IRequestScopeLevel getWormholeScopeToHTTPTools(INKFRequestContext aContext) throws Exception
	{	IKernel kernel=aContext.getKernelContext().getKernel();
		String spaceId="urn:org:netkernel:tpt:http";
		ISpace targetSpace=kernel.getSpace(new SimpleIdentifierImpl(spaceId), null, null);
		if (targetSpace==null)
		{	throw aContext.createFormattedException("EX_SPACE_NOT_FOUND", "MSG_SPACE_NOT_FOUND", null, null, spaceId);
		}
		return RequestScopeLevelImpl.createOrphanedRootScopeLevel(targetSpace, aContext.getKernelContext().getRequestScope());
	}
	
	private void filterLists(ArrayList candidates, ArrayList matches, ArrayList possibles)
	{	candidates.clear();
		for(Object o: matches)
		{	candidates.add(o);
		}
		for(Object o: possibles)
		{	candidates.add(o);
		}
		matches.clear();
		possibles.clear();
	}
	
	private INKFRequest prepareRequest(INKFRequestContext aContext, String aTargetId, String aWrapperTargetId, Map<String, Integer> aVerbTranslate, String method, boolean setOverlayAcceptArgument, String accept) throws Exception
	{	INKFRequest req=aContext.createRequestToEndpoint(aTargetId);
		INKFRequestReadOnly logicalReq=aContext.getThisRequest();
		for(int i=0; i<logicalReq.getArgumentCount(); i++)
		{	String argName=logicalReq.getArgumentName(i);
			if(argName.equals("argType") || argName.equals("scheme"))
			{	break;
			}
			String argValue=logicalReq.getArgumentValue(i);
			req.addArgument(argName, argValue);
		}
		if(setOverlayAcceptArgument)
		{	req.addArgumentByValue("requiredMime", accept);			
		}
		//Wrap the request with the wrapper
		if(aWrapperTargetId!=null)
		{	if(aVerbTranslate!=null)
			{	throw new Exception("Request Wrapping is incompatible with Verb Translation - either don't wrap or don't translate");			
			}
			INKFRequest wrappedReq=req;
			req=aContext.createRequestToEndpoint(aWrapperTargetId);
			WrappedRequestValueSpace vspace=new WrappedRequestValueSpace(wrappedReq);
			vspace.onCommissionSpace(aContext.getKernelContext().getKernel());
			IRequestScopeLevel scope=aContext.getKernelContext().getRequestScope();
			scope=RequestScopeLevelImpl.createDurableScopeLevel(vspace, scope);
			req.setRequestScope(scope);
			req.addArgument("operand", vspace.PBV_IDENTIFIER);
		}
		//VerbTranslation
		if(aVerbTranslate!=null)
		{	Integer verb=aVerbTranslate.get(method);
			if(verb==null)
			{	verb=INKFRequestReadOnly.VERB_SOURCE;				
			}
			if(verb==INKFRequestReadOnly.VERB_SINK  || verb==INKFRequestReadOnly.VERB_NEW)
			{	INKFResponseReadOnly bodyResp=aContext.sourceForResponse("httpRequest:/body");
				req.addPrimaryArgumentFromResponse(bodyResp);
			}
			req.setVerb(verb);
		}
		req.setRepresentationClass(aContext.getThisRequest().getRepresentationClass());
		return req;
	}
		
	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{	return new ISpace[] { mDelegateSpace };
	}

	private void initialize(INKFRequestContext aContext, RESTOverlay aOverlay) throws Exception
	{	ISpaceMeta sm = mDelegateSpaceMeta;
		ISpaceElements sme = sm.getElements();
		Map<String, String> uniqueTargetGrammars=new HashMap();
		Map<String, ArrayList<TargetEndpointData>> uniqueLogicalGrammars=new HashMap();
		boolean errors=false;
		StringBuilder errorBuilder=new StringBuilder();
		for(int i=0; i<sme.size(); i++)
		{	IIdentifier id=sme.getIdentifier(i);
			if(id!=null)
			{	String targetEPID=id.toString();
				IMetaRepresentation meta=aContext.meta(targetEPID);
				IRequestResponseFields state = meta.getAdditionalFields();
				Object value = state.getValue("meta");
				if (value instanceof Element)
				{	IHDSNode n = HDSFactory.parseDOM((Element) value);
					n=n.getFirstNode("/meta/rest");
					if(n!=null)
					{	String simple=(String)n.getFirstValue("simple");
						if(simple==null)
						{	simple="";						
						}
						
						String grammar=mBasePath+simple;
						String aIdentifier;
						String eid=(String)n.getFirstValue("id");
						if(eid==null)
						{	aIdentifier=constructIdentifier(grammar);				
						}
						else
						{	aIdentifier=eid;
						}
						
						LogicalEndpointData led=mLogicalEndpoints.get(aIdentifier);
						if(led==null)
						{	led=new LogicalEndpointData(aIdentifier, grammar, aContext.getKernelContext().getKernel().getLogger(), state);
						}
						//Check that target endpoint grammar is unique and log a warning if it is not
						if(meta instanceof IEndpointMeta)
						{	IEndpointMeta epm=(IEndpointMeta)meta;
							String gs=epm.getIdentifierGrammar().toString();
							String existingGrammar=uniqueTargetGrammars.get(gs);
							if(existingGrammar!=null)
							{	aContext.logFormatted(INKFRequestContext.LEVEL_WARNING, "WARNING_RESTOVERLAY_AMBIGUOUSGRAMMAR", existingGrammar, targetEPID);
								errors=true;
								errorBuilder.append("Identical Internal Grammar - the target endpoint ["+targetEPID+"]  has a non-unique grammar ["+existingGrammar+"]\n");
							}
							else
							{	uniqueTargetGrammars.put(gs, gs);								
							}
						}
						TargetEndpointData ted=new TargetEndpointData(targetEPID, simple, n);
						led.mMappings.add(ted);
						mLogicalEndpoints.put(led.mIdentifier, led);
						//Also provide bidirectional map from grammar to targets
						ArrayList<TargetEndpointData> targetList=mGrammarToTargets.get(grammar);
						if(targetList==null)
						{ targetList=new ArrayList<TargetEndpointData>();	
						}
						targetList.add(ted);
						mGrammarToTargets.put(grammar, targetList);
						//Test for degeneracy and warn about it - 
						String normalisedSimple=simple.replaceAll("\\{.*?\\}", "*");
						boolean validTed=true;
						ArrayList<TargetEndpointData> testarray=uniqueLogicalGrammars.get(normalisedSimple);
						if(testarray!=null)
						{	for(TargetEndpointData testted : testarray)
							{	if(testted.mGrammar.equals(ted.mGrammar))
								{	if(testted.mMethod==null || (testted.mMethod !=null && testted.mMethod.equals(ted.mMethod)))
									{	if(testted.mConsumes==null || (testted.mConsumes !=null && testted.mConsumes.equals(ted.mConsumes)))
										{	if(testted.mProduces==null || (testted.mProduces !=null && testted.mProduces.equals(ted.mProduces)))
											{	if(testted.mLanguage==null || (testted.mLanguage !=null && testted.mLanguage.equals(ted.mLanguage)))
												{		//Error this is identical declaration.
														aContext.logFormatted(INKFRequestContext.LEVEL_WARNING, "WARNING_RESTOVERLAY_IDENTICALGRAMMAR", targetEPID, ted.mGrammar);
														errors=true;
														errorBuilder.append("Identical External Grammar - the target endpoint ["+targetEPID+"]  has an identical grammar ["+ted.mGrammar+"] and all demux fields are degenerate\n");
												}
											}
										}
									}
								}
								else
								{	if(n.getFirstNode(IGNORE_AMBIGUITY_WARNING)==null)
									{	if(testted.mMethod==null || (testted.mMethod !=null && testted.mMethod.equals(ted.mMethod)))
										{	if(testted.mConsumes==null || (testted.mConsumes !=null && testted.mConsumes.equals(ted.mConsumes)))
											{	if(testted.mProduces==null || (testted.mProduces !=null && testted.mProduces.equals(ted.mProduces)))
												{	if(testted.mLanguage==null || (testted.mLanguage !=null && testted.mLanguage.equals(ted.mLanguage)))
													{	StringBuilder sb=new StringBuilder();
														for(TargetEndpointData lted : testarray)
														{	sb.append(lted.mGrammar).append("  ");													
														}
														//Warn this is not an exception since its possible that the grammars are distinct
														aContext.logFormatted(INKFRequestContext.LEVEL_WARNING, "WARNING_RESTOVERLAY_DEGENERATELOGICALENDPOINT", targetEPID, simple, sb.toString());
														if(mStrict)
														{	//In strict mode so mark space as broken.	
															errors=true;
															errorBuilder.append("Ambiguous External Grammar endpoint ["+targetEPID+"] has potentially ambiguous simple grammar ["+ted.mGrammar+"] overlapping ["+sb.toString()+"] \n");
														}
													}
												}
											}
										}
									}
									else
									{	aContext.logFormatted(INKFRequestContext.LEVEL_INFO, "INFO_RESTOVERLAY_DEGENERATELOGICALENDPOINT", targetEPID, ted.mGrammar);
									}
								}
							}
						}
						else
						{	testarray=new ArrayList();							
						}
						if(validTed)
						{	testarray.add(ted);
							uniqueLogicalGrammars.put(normalisedSimple, testarray);
						}
					}
				}
				else
				{	//Transparently relay through any non-RESTOverlay managed endpoint
					if(meta instanceof IEndpointMeta)
					{	//String proxyID=constructIdentifier(targetEPID);
						mLogicalEndpoints.put(targetEPID, new LogicalEndpointData(targetEPID, meta));
						mPassThroughEndpoints.put(targetEPID, targetEPID);
					}
				}
			}
			if(errors)
			{	aOverlay.setInitialisationProblem(errorBuilder.toString());
			}
		}
		// Construct space elements with a tuple list {IIdentifier aIndentifier, ISpace aDelegatedSpace, IEndpoint aPhysicalEndpoint}
		int tlSize=0;
		if(mLogicalEndpoints.size()!=0)
		{	tlSize=mLogicalEndpoints.size();			
		}
		else if(mAuto404!=null)
		{	tlSize=1;			
		}
		TupleList tl=new TupleList(3, tlSize);
		for(String key: mLogicalEndpoints.keySet())
		{		tl.put(new SimpleIdentifierImpl(key), null, mRESTOverlay);
		}
		if(mAuto404!=null)
		{		String grammar=mBasePath+"{404}";
				mAuto404EndpointIdentifier=constructIdentifier(grammar);
				LogicalEndpointData led=new LogicalEndpointData(mAuto404EndpointIdentifier, grammar, aContext.getKernelContext().getKernel().getLogger(), RequestResponseFieldsImpl.EMPTY);
				mLogicalEndpoints.put(led.mIdentifier, led);
				tl.put(new SimpleIdentifierImpl(mAuto404EndpointIdentifier), null, mRESTOverlay);
		}
		
		mSpaceElements=new SpaceElementsImpl(tl);
	}
	
	private class LogicalEndpointData
	{	public String mIdentifier;
		public ArrayList<TargetEndpointData> mMappings;
		public IMetaRepresentation mEndpointMeta;
		public String mGrammar;		//Required so that once resolution to LogicalEP has occured we can go back to the unresolved set of potential targets in mGrammarToTargets
		
		public LogicalEndpointData(String aIdentifier, IMetaRepresentation aEndpointMeta)
		{	mEndpointMeta=aEndpointMeta;
			mIdentifier=aIdentifier;
		}
		
		public LogicalEndpointData(String aIdentifier, String aGrammar, IMessages aMessages, IRequestResponseFields aFields) throws Exception
		{	mIdentifier=aIdentifier;
			constructMetadata(aGrammar, aMessages, aFields);
			mGrammar=aGrammar;
			mMappings=new ArrayList();
		}
		
		private void constructMetadata(String aGrammar, IMessages aMessages, IRequestResponseFields aFields) throws Exception
		{	String grammarImpl="res:"+aGrammar;
			String grammar="<grammar><simple>"+grammarImpl+"</simple></grammar>";
			Document gd=XMLUtils.parse(new ByteArrayInputStream(grammar.getBytes()));
			//IIdentifierGrammar g=BNFGrammarFactory.parse(gd.getDocumentElement(), aMessages);
			EndpointMetaBuilder epm=new EndpointMetaBuilder();
			MetadataUtils.parseGrammar(gd.getDocumentElement(), epm, aMessages);
			IEndpointMeta meta=epm.getMeta(aMessages);
			ClonableEndpointMetaImpl meta2=new ClonableEndpointMetaImpl(
					grammarImpl,
					"Auto-generated RESTOverlay logical endpoint",
					null,
					null,
					INKFRequestReadOnly.VERB_SOURCE,
					meta.getIdentifierGrammar(),
					EndpointMetaImpl.NO_ARGS,
					EndpointMetaImpl.NO_RESPONSES,
					EndpointMetaImpl.NO_EXCEPTIONS
				);
			meta2.setAdditionalFields(aFields);
			mEndpointMeta=meta2;
		}
	}
	
	private class TargetEndpointData
	{	public String mIdentifier;
		public IHDSNode mMeta;
		public String mMethod;
		public Map<String, IHDSNode> mProduces;
		public String mConsumes;
		public String mLanguage;
		public boolean mCompress;
		public boolean mMustBeSSL;
		public String mExceptionTarget;
		public String mPreTarget;
		public String mWrapperTarget;
		public String mEtag;
		public int mMaxContentLength=-1;
		public Map<String, Integer> mVerbTranslate;
		public String mGrammar;
		
		public TargetEndpointData(String aIdentifier, String aGrammar, IHDSNode aMeta)
		{	mIdentifier=aIdentifier;
			mGrammar=aGrammar;
			mMeta=aMeta;
			analyseMeta();
		}
		
		private void analyseMeta()
		{	String method=(String)mMeta.getFirstValue("method");
			if(method!=null)
			{	method=method.toUpperCase();				
			}
			mMethod=method;
			//mProduces=(String)mMeta.getFirstValue("produces");
			IHDSNodeList nl=mMeta.getNodes("produces");
			if(nl.size()>0)
			{	mProduces=new HashMap(nl.size());
				for(IHDSNode n: nl)
				{	String v=(String)n.getValue();
					mProduces.put(v, n);
				}
			}
			mConsumes=(String)mMeta.getFirstValue("consumes");
			mLanguage=(String)mMeta.getFirstValue("language");
			mCompress=mMeta.getFirstNode("compress")!=null;
			mMustBeSSL=mMeta.getFirstNode("mustBeSSL")!=null;
			mExceptionTarget=(String)mMeta.getFirstValue("exceptionTarget");
			mPreTarget=(String)mMeta.getFirstValue("preTarget");
			mWrapperTarget=(String)mMeta.getFirstValue("wrapperTarget");
			mEtag=(String)mMeta.getFirstValue("Etag");
			if(mEtag!=null)
			{	mEtag=mEtag.toUpperCase();				
			}
			String maxContentLength=(String)mMeta.getFirstValue("maxContentLength");
			if(maxContentLength!=null)
			{	mMaxContentLength=Integer.parseInt(maxContentLength);
			}
			boolean doVerbTranslate=mMeta.getFirstNode("verbTranslate")!=null;
			if(doVerbTranslate)
			{	mVerbTranslate=new HashMap();
				mVerbTranslate.put("GET", INKFRequestReadOnly.VERB_SOURCE);
				mVerbTranslate.put("POST", INKFRequestReadOnly.VERB_SINK);
				mVerbTranslate.put("PUT", INKFRequestReadOnly.VERB_SINK);
				mVerbTranslate.put("DELETE", INKFRequestReadOnly.VERB_DELETE);
				String verbTranslate=(String)mMeta.getFirstValue("verbTranslate");
				if(verbTranslate!=null)
				{	Class c=INKFRequestReadOnly.class;
					verbTranslate=verbTranslate.toUpperCase();
					String[] s1=verbTranslate.split(",");
					for(int i=0; i<s1.length; i++)
					{	String s=s1[i].trim();
						String[] s2=s.split(":");
						String m=s2[0];
						String v=s2[1];
						try
						{	Field f=c.getField("VERB_"+v);
							mVerbTranslate.put(m, f.getInt(null));
						}
						catch(Exception e)
						{	//Do nothing							
						}
						
					}
				}
			}
		}
	}
	
	private class ClonableEndpointMetaImpl extends MetaRepresentationImpl implements IEndpointMeta
	{
		private final int mVerbs;
		private final IIdentifierGrammar mGrammar;
		private final IEndpointArgumentMeta[] mParts;
		private final Class[] mResponseClasses;
		private final IMetaRepresentation[] mExceptions;
		
		public ClonableEndpointMetaImpl(String aName, String aDescription, IIdentifier aIcon, IIdentifier aDoc,
				int aVerbs, IIdentifierGrammar aGrammar, IEndpointArgumentMeta[] aParts, Class[] aResponseClasses, IMetaRepresentation[] aExceptions)
		{	super(aName,aDescription,aIcon,aDoc);
			mVerbs=aVerbs;
			mGrammar=aGrammar;
			mParts=aParts;
			mResponseClasses=aResponseClasses;
			mExceptions = aExceptions;
		}

		public int getSupportedVerbs()
		{	return mVerbs;
		}

		public IIdentifierGrammar getIdentifierGrammar()
		{	return mGrammar;
		}

		public IEndpointArgumentMeta[] getArguments()
		{	return mParts;
		}

		public Class[] getResponseClasses()
		{	return mResponseClasses;
		}

		public IMetaRepresentation[] getExceptions()
		{	return mExceptions;
		}
		
		public void setAdditionalFields(IRequestResponseFields aFields)
		{	super.setAdditionalFields(aFields);
		}
	}
	
	public String constructIdentifier(String aId)
	{	return aId.replaceAll("/", "_").replaceAll("\\?", "_");
		//return Utils.generateUniqueEndpointIdentifier(mRESTOverlay.getOwningSpace(), mRESTOverlay, aId);
	}
}
