/******************************************************************************
  (c) Copyright 2002-2008, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      

  File:          $RCSfile: HTTPSessionContext.java,v $
  Version:       $Name:  $ $Revision: 1.10 $
  Last Modified: $Date: 2008/04/16 10:08:18 $
 *****************************************************************************/
package org.netkernel.http.session;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFSpaceImpl;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.layer0.util.TimedExpiryFunction;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.urii.ClassLoaderWithExports;
import org.netkernel.urii.IEndpoint;


/** Context and Endpoint to fulfill requests for session: address space
 * @author pjr
 *
 */
public class HTTPSessionSpace extends StandardSpace
{	
	private static IEndpoint sEndpoint = new HTTPSessionEndpoint();
	private static final String SESSION_SCHEME="session:";
	private GoldenThreadExpiryFunction mExpiryFunction;
	private final String mId;
	
	private Map mSession;

	public HTTPSessionSpace(StandardSpace aOwningSpace, HTTPSessionSpace2Proto aProtoSpace, String aId)
	{	super(aOwningSpace, aOwningSpace.getOwningModule(),null);
		setMappings(aProtoSpace.getEndpoints());
		setName("HTTP Session Space");
		setState(ISpaceMeta.STATE_PROTOSPACE, aProtoSpace );
		
		mSession=new ConcurrentHashMap();
		mExpiryFunction=new GoldenThreadExpiryFunction("HTTPSessionExpiryFunction");
		mId=aId;
	}
	
	public String toString()
	{	return "HTTP Session Space ["+mId+"]";
	}

	public void handleRequestFromEndpoint(INKFRequestContext helper) throws Exception
	{	String identifier=helper.getThisRequest().getIdentifier().substring(SESSION_SCHEME.length());

		//Process session request
		Object response=null;
		switch(helper.getThisRequest().getVerb())
		{	case INKFRequestReadOnly.VERB_SOURCE:
				response=mSession.get(identifier);
			break;
			case INKFRequestReadOnly.VERB_SINK:
				mSession.put(identifier,helper.getThisRequest().getPrimary());
				setDirty();
			break;
			case INKFRequestReadOnly.VERB_DELETE:
				if(identifier.length()==0)
				{	//Delete everything
					mSession.clear();
				}
				else
				{	mSession.remove(identifier);
				}
				response=Boolean.TRUE;
				setDirty();
			break;
			case INKFRequestReadOnly.VERB_EXISTS:
				response= mSession.containsKey(identifier) ? Boolean.TRUE : Boolean.FALSE;
			break;
		}
		INKFResponse resp=helper.createResponseFrom(response);
		resp.setExpiry(INKFResponse.EXPIRY_FUNCTION, mExpiryFunction);
	}
	
	public int getItemCount()
	{	return mSession.size();		
	}
	
	synchronized void setDirty()
	{	mExpiryFunction.invalidate();
		mExpiryFunction=new GoldenThreadExpiryFunction("HTTPSessionExpiryFunction");
	}
	
	public boolean resolveEquals(Object aOther)
	{	return aOther instanceof HTTPSessionSpace;
	}

	@Override
	public ClassLoaderWithExports getClassLoader()
	{	return null;
	}
}