package org.netkernel.http.session;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.http.transport.HTTPBridgeOverlay;
import org.netkernel.http.transport.HTTPRequestSpace2Wrapper;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.impl.EndpointArgumentMetaImpl;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.module.standard.endpoint.StandardEndpointImpl;
import org.netkernel.urii.ISpace;

public class HTTPSessionSpace2Proto extends StandardSpace
{
private final IStandardEndpoint[] mEndpoints;
	
	public HTTPSessionSpace2Proto(StandardSpace aOwningSpace, HTTPSessionOverlay aOverlay, IKernel aKernel) throws Exception
	{	super(aOwningSpace,aOwningSpace.getOwningModule(),null);
		IStandardEndpoint ep1 = new HTTPSessionEndpoint(aKernel);
		IStandardEndpoint ep2 = new HTTPSessionMetadataEndpoint(aOverlay,aKernel);
		mEndpoints = new IStandardEndpoint[]{ep1,ep2};	
		setMappings(mEndpoints);
		setName("HTTP Session Space");
	}
	
	public IStandardEndpoint[] getEndpoints()
	{	return mEndpoints;
	}
	
	private static class HTTPSessionEndpoint extends StandardAccessorImpl
	{
		public HTTPSessionEndpoint(IKernel aKernel) throws Exception
		{	
			this.declareName("SessionMetadataEndpoint");
			this.declareDescription("Provides statistics for the HTTP sessions");
			this.declareThreadSafe();
			declareName("HTTPSessionEndpoint");
			IGrammarGroup g1=BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createText(g1, "session:");
			IGrammarGroup g2=BNFGrammarFactory.createGroup(g1,"path",null);
			BNFGrammarFactory.createRegex(g2, BNFGrammarFactory.BUILTIN_ANYTHING);			
			ILogger logger=aKernel.getLogger();
			IIdentifierGrammar g=BNFGrammarFactory.createGrammar(g1,logger);
			declareGrammar(g);
			declareArgument(new EndpointArgumentMetaImpl("path", "a unique resource path for storing session state", null, null, null, false, null));
			//declareSupportedVerbs(INKFRequestReadOnly.VERB_SOURCE|INKFRequestReadOnly.VERB_SINK|INKFRequestReadOnly.VERB_EXISTS|INKFRequestReadOnly.VERB_DELETE);
			setParameterRaw("id", "http:session", aKernel);
		}
		
		public void onSource(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPSessionSpace)
			{	((HTTPSessionSpace)space).handleRequestFromEndpoint(aContext);
			}
		}		
		public void onSink(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPSessionSpace)
			{	((HTTPSessionSpace)space).handleRequestFromEndpoint(aContext);
			}
		}		
		public void onExists(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPSessionSpace)
			{	((HTTPSessionSpace)space).handleRequestFromEndpoint(aContext);
			}
		}		
		public void onDelete(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPSessionSpace)
			{	((HTTPSessionSpace)space).handleRequestFromEndpoint(aContext);
			}
		}		
	}	
}
