package org.netkernel.http.transport;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.impl.EndpointArgumentMetaImpl;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.ISpace;

public class HTTPRequestSpace2 extends StandardSpace
{
	private final IStandardEndpoint[] mEndpoints;
	
	public HTTPRequestSpace2(HTTPBridgeOverlay aBridge, IKernel aKernel) throws Exception
	{	super(aBridge.getOwningSpace(),aBridge.getOwningSpace().getOwningModule(),null);
		mEndpoints = createEndpoints(aKernel);
		setMappings(mEndpoints);
		setName("HTTP Request Response Space");
	}
	
	private static IStandardEndpoint[] createEndpoints(IKernel aKernel) throws Exception
	{	IStandardEndpoint ep1 = new HTTPRequestEndpoint1(aKernel);
		IStandardEndpoint ep2 = new HTTPRequestEndpoint2(aKernel);
		IStandardEndpoint ep3 = new HTTPResponseEndpoint(aKernel);
		IStandardEndpoint ep4 = new WSTransport(aKernel);
		return new IStandardEndpoint[]{ep1,ep2,ep3,ep4};
	}
	
	public IStandardEndpoint[] getEndpoints()
	{	return mEndpoints;
	}
	
	private static class HTTPRequestEndpoint1 extends HTTPRequestEndpoint
	{	public HTTPRequestEndpoint1(IKernel aKernel) throws Exception
		{	IGrammarGroup g1=BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createText(g1, "httpRequest:/");
			IGrammarGroup g2=BNFGrammarFactory.createGroup(g1,"type",null);
			BNFGrammarFactory.createRegex(g2, BNFGrammarFactory.BUILTIN_PATH_SEGMENT);
			ILogger logger=aKernel.getLogger();
			IIdentifierGrammar g=BNFGrammarFactory.createGrammar(g1,logger);
			declareGrammar(g);
			declareArgument(new EndpointArgumentMetaImpl("type", "type of request resource", null, null, null, false, null));
			setParameterRaw("id", "http:request1", aKernel);
		}
	}
	private static class HTTPRequestEndpoint2 extends HTTPRequestEndpoint
	{	public HTTPRequestEndpoint2(IKernel aKernel) throws Exception
		{	IGrammarGroup g1=BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createText(g1, "httpRequest:/");
			IGrammarGroup g2=BNFGrammarFactory.createGroup(g1,"type",null);
			BNFGrammarFactory.createRegex(g2, BNFGrammarFactory.BUILTIN_PATH_SEGMENT);
			BNFGrammarFactory.createText(g1, "/");
			IGrammarGroup g3=BNFGrammarFactory.createGroup(g1,"name",null);
			BNFGrammarFactory.createRegex(g3, BNFGrammarFactory.BUILTIN_ANYTHING);
			ILogger logger=aKernel.getLogger();
			IIdentifierGrammar g=BNFGrammarFactory.createGrammar(g1,logger);
			declareGrammar(g);
			declareArgument(new EndpointArgumentMetaImpl("type", "type of request resource", null, null, null, false, null));
			declareArgument(new EndpointArgumentMetaImpl("name", "name of request resource", null, null, null, false, null));
			setParameterRaw("id", "http:request2", aKernel);
		}
	}

	private static class HTTPRequestEndpoint extends StandardAccessorImpl
	{
		public HTTPRequestEndpoint()
		{	declareThreadSafe();
		}
		
		public void onSource(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPRequestSpace2Wrapper)
			{	((HTTPRequestSpace2Wrapper)space).onRequestFromEndpoint(aContext);
			}
		}
		
		public void onExists(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPRequestSpace2Wrapper)
			{	((HTTPRequestSpace2Wrapper)space).onRequestFromEndpoint(aContext);
			}
		}
	}

	private static class HTTPResponseEndpoint extends StandardAccessorImpl
	{
	
		public HTTPResponseEndpoint(IKernel aKernel) throws Exception
		{	this.declareThreadSafe();
			IGrammarGroup g1=BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createText(g1, "httpResponse:/");
			IGrammarGroup g2=BNFGrammarFactory.createGroup(g1,"path",null);
			BNFGrammarFactory.createRegex(g2, BNFGrammarFactory.BUILTIN_ANYTHING);			
			ILogger logger=aKernel.getLogger();
			IIdentifierGrammar g=BNFGrammarFactory.createGrammar(g1,logger);
			declareGrammar(g);
			setParameterRaw("id", "http:response", aKernel);
		}
		
		public void onSink(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPRequestSpace2Wrapper)
			{	((HTTPRequestSpace2Wrapper)space).onRequestFromEndpoint(aContext);
			}
		}
	}

	private static class WSTransport extends StandardAccessorImpl
	{
		public WSTransport(IKernel aKernel) throws Exception
		{	declareThreadSafe();
			IGrammarGroup g1=BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createText(g1, "wsResponse:/");
			IGrammarGroup g2=BNFGrammarFactory.createGroup(g1,"path",null);
			BNFGrammarFactory.createRegex(g2, "message|disconnect");			
			ILogger logger=aKernel.getLogger();
			IIdentifierGrammar g=BNFGrammarFactory.createGrammar(g1,logger);
			declareGrammar(g);
			setParameterRaw("id", "ws:response", aKernel);
		}
		
		public void onSink(INKFRequestContext aContext) throws Exception
		{	ISpace space=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			if (space instanceof HTTPRequestSpace2Wrapper)
			{	((HTTPRequestSpace2Wrapper)space).onRequestFromEndpoint(aContext);
			}
		}
	}
}
