package org.netkernel.http.transport;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.concurrent.atomic.AtomicInteger;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;


import org.eclipse.jetty.http.HttpHeader;
import org.eclipse.jetty.server.Request;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.AbstractHandler;
import org.eclipse.jetty.server.handler.ContextHandler;
import org.eclipse.jetty.server.handler.ContextHandler.Context;
import org.netkernel.http.transport.representation.HTTPRequestResponseRepresentation;
import org.netkernel.layer0.nkf.INKFLocale;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.util.Utils;
import org.netkernel.urii.INetKernelThrowable;

public class NetKernelHandler extends AbstractHandler implements INetKernelHandler 
{
	private INKFRequestContext mContext;
	private AtomicInteger mRequestCount=new AtomicInteger();
	private Boolean mUsePriorityRequests=false;
	
	public NetKernelHandler()
	{
	}
	
	@Override
	public void setServer(Server aServer)
	{	super.setServer(aServer);
		NKContextServer server=(NKContextServer)getServer();
		mContext=server.getTransportContext();
		server.registerNetKernelHandler(this);
		mContext.logFormatted(INKFLocale.LEVEL_INFO,"MSG_REGISTERED_SYNC_HANDLER", new Object[] {null});
	}
	
	public void destroy()
	{	super.destroy();
		mContext=null;
	}
	
		public void handle(String target, Request lowLevelJettyRequest, HttpServletRequest request,	HttpServletResponse response) throws IOException, ServletException
	{	
		try
		{	mRequestCount.incrementAndGet();
			ContextHandler ch=new  ContextHandler("/");
			ch.setBaseResource(new DummyContextResource());
			Context context=ch.getServletContext();
			lowLevelJettyRequest.setContext(context);
			HTTPRequestResponseRepresentation rep = new HTTPRequestResponseRepresentation(request,response, null);
			String url=Utils.decode(request.getRequestURL().toString());
			String qs=request.getQueryString();
			if (qs!=null)
			{	url=url+"?"+qs;
			}
			INKFRequest req = mContext.createRequest(url);
			req.addPrimaryArgument(rep);
			if(mUsePriorityRequests)
			{	req.setPriority(INKFRequestReadOnly.PRIORITY_MAX);		
			}
			mContext.issueRequestForResponse(req);
		}
		catch (Throwable t)
		{	response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
			response.setDateHeader(HttpHeader.EXPIRES.asString(),0);
			response.setContentType("text/plain");
			response.setHeader(HttpHeader.SERVER.asString(), HTTPTransportEndpoint.getServerHeader() );
			OutputStream os=response.getOutputStream();
			PrintWriter pw = new PrintWriter(os);
			if (t instanceof INetKernelThrowable)
			{	pw.write(t.toString());
			}
			else
			{	t.printStackTrace(pw);
			}
			pw.flush();
		}
		finally
		{	response.flushBuffer();
			lowLevelJettyRequest.setHandled(true);
			lowLevelJettyRequest.setContext(null);
		}
	}
	
	public void setPriorityRequests(boolean b)
	{	mUsePriorityRequests=b;		
	}

	public int requestCount()
	{	return mRequestCount.get();
	}
	
}
